version='2.8.17'
version_cmd='redis-server -v'
depends=['build-essential']
download_url='http://download.redis.io/releases/redis-VERSION.tar.gz'
install_script="""
tar -xzf redis-VERSION.tar.gz
cd redis-VERSION/
make
make install
"""


version='4.3.20'
version_cmd='VBoxManage -v'
download_url='http://download.virtualbox.org/virtualbox/4.3.20/virtualbox-4.3_4.3.20-96996~Ubuntu~raring_amd64.deb'
install_script="""
dpkg -i virtualbox-4.3_4.3.20-96996~Ubuntu~raring_amd64.deb || true
apt-get install -fy
"""


version='0.4.0'
version_cmd='consul -v'
depends=['unzip']
download_url='https://dl.bintray.com/mitchellh/consul/VERSION_linux_amd64.zip'
install_script="""
unzip -o VERSION_linux_amd64.zip
mv -f consul /usr/local/bin/consul
"""


name='mongodb'
version='2.6.7'
version_cmd='mongod --version'
install_script='''
apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv 7F0CEB10
echo 'deb http://downloads-distro.mongodb.org/repo/ubuntu-upstart dist 10gen' > /etc/apt/sources.list.d/mongodb.list
apt-get update -qq
apt-get install -qy mongodb-org
'''


version='1.2.0'
version_cmd='docker -v'
depends=['apt-transport-https']
install_script="""
echo deb https://get.docker.io/ubuntu docker main > /etc/apt/sources.list.d/docker.list
apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 --recv-keys 36A1D7869245C8950F966E92D8576A8BA88D21E9
apt-get update
apt-get install -y lxc-docker
"""


import sys
import datetime
from textwrap import TextWrapper, dedent


import click
from colorama import Style, Fore
B = Style.BRIGHT
N = Style.NORMAL


import lib


def get_app(app_name):
    app = lib.App.load(app_name)
    if app is None:
        click.echo('ifs does not have a source for "%s"' % app_name)
        exit(1)
    else:
        return app

def error(message=None, nl=True):
    click.echo(Style.BRIGHT + Fore.RED + message + Style.RESET_ALL, sys.stderr, nl)

def ok(message=None, file=None, nl=True):
    click.echo(Style.BRIGHT + Fore.GREEN + message + Style.RESET_ALL, file, nl)


@click.group()
@click.version_option()
def cli():
    """
    Install From Source

    When the version in the package manager has gone stale, get a fresh,
    production-ready version from a source tarball or precompiled archive.

    Each installation script is called a `source`, identified by a name like
    `nginx`. ifs commands generally operate on these sources e.g. `ifs install
    nginx` or `ifs export nginx`.

    For a list of commands type `ifs ls`. For additional help on any command
    type `ifs COMMAND --help`, e.g. `ifs install --help`.

    Please send issues or improvements to https://github.com/cbednarski/ifs
    """
    pass


@cli.command()
def ls():
    """
    List available sources.

      ifs ls
    """
    for app in lib.list_app_names():
        click.echo(app)


@cli.command()
@click.argument('term')
def grep(term):
    """
    Search in source name and description.

      ifs search nginx
    """
    apps = lib.search_app_strings(term)
    text = TextWrapper(initial_indent='  ', subsequent_indent='  ')
    for app in apps:
        click.echo(app.name)
        if app.description:
            des = dedent(app.description)
            click.echo(text.fill(app.description))


@cli.command()
@click.argument('app_name')
@click.option('--version', help='Override the default version')
@click.option('--force', '-f', default=False, is_flag=True, help='Force redownload and reinstall')
def install(app_name, version, force):
    """
    Install the specified application.

      ifs install -f --version 1.6.2 nginx

    To inspect an install script before you run it, use `ifs export SOURCE`.

    Versions

    Sources come in three varieties: compiled packages, source packages,
    and repositories. If a source uses a package (tarball, .deb) the install
    script will be locked to a version. You can override it with --version

      ifs install redis --version 2.8.13

    In this case, if you want to upgrade to a new version you can generally
    just specify the latest version when it comes out, e.g.

      ifs install redis --version 2.8.15

    In the third case packages are installed from an apt repo. By default, ifs
    will install the latest version from the repo when you run install. If you
    want to upgrade later, you can pass -f, e.g. `ifs install -f jenkins`. This
    will force ifs to reinstall the latest (essentially upgrading).

    Upstream Patches

    We strive to provide installers for the latest and greatest versions.

    Sometimes the installation dependencies will change and you'll need to
    modify the installer to get the latest version to install. Please submit a
    patch if you fix something! https://github.com/cbednarski/ifs

    Future-proofing Installations

    There are a slew of variables involved in installing packages, including
    updates to dependencies, security fixes, and so forth.

    If you want to guarantee that the version you install doesn't change over
    time, you should probably use a tool like docker. However, you can achieve
    some naive version-locking by using `ifs export SOURCE` for each app you
    install, and checking these files into your version control. When you
    install later, use `ifs source FILENAME` instead of `ifs install SOURCE`.
    """
    if app_name == 'template':
        click.echo("`template` isn't really an app. It's a dummy that helps "
                   "you create a new ifs installer. Try:\n  ifs export "
                   "template")
        exit(1)

    app = get_app(app_name)
    installed_version = app.check_version()

    if not version:
        version = app.version
    if version and installed_version == version and not force:
        click.echo('%s %s is already installed' % (app.name, installed_version))
        exit(0)
    if installed_version and not version and not force:
        click.echo('%s %s is already installed' % (app.name, installed_version))
        exit(0)

    cmd = app.install(version)
    if cmd.returncode == 0:
        ok(cmd.output)
    else:
        error(cmd.output)
    exit(cmd.returncode)


@cli.command()
@click.argument('source', type=click.Path(exists=True))
def source(source):
    """
    Install from a custom source.

      ifs source /path/to/source.py

    ifs ships with some built-in source files. This command allows you to load
    your own, arbitrarily-defined source file at runtime. Think of it as the
    bash source built-in for ifs.

    Note: The source file should be a python file with the same basic structure
    as one of the built-in source files. Run `ifs export template` to create a
    new one.
    """
    app = lib.App.load_external(source)
    if app:
        cmd = app.install()
        if cmd.returncode == 0:
            ok('Installation succeeded')
        else:
            error('Installation failed')
        exit(cmd.returncode)
    else:
        error('Unable to load source from %s' % source)
        exit(1)


@cli.command()
@click.argument('source')
def export(source):
    """
    Export a built-in source.

      ifs export template > newapp.py

    This is useful to see details on how a built-in source is installed, or to
    facilitate creating a new source from a template. You can install exported
    sources with `ifs source FILENAME`.
    """
    source_code = lib.export_source(source)
    if source_code:
        click.echo(source_code)
    else:
        error('No built-in source for %s' % source)
        exit(1)


@cli.command()
@click.argument('app_name')
def version(app_name):
    """
    Show the installed source version.

      ifs version nginx

    Abstracts away application-specific flags like -v, -V, --version, etc.
    """
    app = get_app(app_name)

    version = app.check_version()
    if version is None:
        error('%s is not installed' % app_name)
        exit(1)

    click.echo(version)


@cli.command()
@click.argument('app_name')
def info(app_name):
    """
    Show source name, version, and deps.

      ifs info nginx

    Run `ifs ls` for a list of applications
    """
    app = get_app(app_name)
    info = app.info()
    for k, v in info.iteritems():
        if type(v) is list:
            v = ' '.join(v)
        click.echo('%s: %s' % (k, v))


@cli.command()
@click.option('--force', '-f', default=False, is_flag=True, help='Force update, even if the system was updated recently')
def aptupdate(force):
    """
    Alias for apt-get upgrade and associated housekeeping (clean, autoremove, etc.).
    """
    if force or lib.out_of_date():
        cmd = lib.Cmd.run("""
            apt-get clean
            apt-get update -qq
            apt-get upgrade -yq
            apt-get autoremove -y
            """, autoprint=True)
        lib.write_last_updated()
        exit(cmd.returncode)
    else:
        t = datetime.datetime.fromtimestamp(lib.read_last_updated()).isoformat()
        click.echo("System was updated at %s, skipping" % t)
    exit(0)


if __name__ == '__main__':
    cli()



version='0.4.6'
version_cmd='etcd -version'
download_url='https://github.com/coreos/etcd/releases/download/vVERSION/etcd-vVERSION-linux-amd64.tar.gz'
install_script="""
tar -xzf etcd-vVERSION-linux-amd64.tar.gz
mv -f ./etcd-v0.4.6-linux-amd64/etcd /usr/local/bin/etcd
mv -f ./etcd-v0.4.6-linux-amd64/etcdctl /usr/local/bin/etcdctl
"""



version='0.10.31'
version_cmd='node -v'
download_url='http://nodejs.org/dist/vVERSION/node-vVERSION-linux-x64.tar.gz'
install_script="""
  tar --strip-components 1 -C /usr/local -xzf node-vVERSION-linux-x64.tar.gz
"""


version_cmd='mysqld --version'
version_re='(\d+\.\d+\.\S+).+Percona Server'
install_script="""
apt-key adv --keyserver keys.gnupg.net --recv-keys 1C4CBDCDCD2EFD2A
echo "deb http://repo.percona.com/apt trusty main
deb-src http://repo.percona.com/apt trusty main" > /etc/apt/sources.list.d/percona.list
apt-get update -qq
DEBIAN_FRONTEND=noninteractive apt-get install -y percona-server-server percona-server-client

mysql -e "CREATE FUNCTION fnv1a_64 RETURNS INTEGER SONAME 'libfnv1a_udf.so'" || true
mysql -e "CREATE FUNCTION fnv_64 RETURNS INTEGER SONAME 'libfnv_udf.so'" || true
mysql -e "CREATE FUNCTION murmur_hash RETURNS INTEGER SONAME 'libmurmur_udf.so'" || true
"""


version='1.4.2'
version_cmd='logstash version'
version_re='(\d+\.\d+\.\S+)'
download_url='https://download.elasticsearch.org/logstash/logstash/packages/debian/logstash_1.4.2-1-2c0f5a1_all.deb'
install_script="""
dpkg -i logstash_1.4.2-1-2c0f5a1_all.deb
"""



version='1.5.5'
version_cmd='haproxy -v'
depends=['build-essential','libpcre3-dev','zlib1g-dev','libssl-dev']
download_url='http://www.haproxy.org/download/1.5/src/haproxy-VERSION.tar.gz'

upstart="""
start on runlevel [2345]
stop on runlevel [016]

expect fork

pre-start exec haproxy -t
pre-stop exec haproxy -s quit

reload signal SIGHUP

exec haproxy
"""

install_script="""
echo "This installer is broken :("
echo "Please help fix it: https://github.com/cbednarski/ifs/issues/9"
# tar -xzf haproxy-VERSION.tar.gz
# cd haproxy-VERSION
# make TARGET=generic USE_TFO=1 USE_PCRE=1 USE_STATIC_PCRE=1 USE_PCRE_JIT=1 USE_OPENSSL=1
# make install
# if [ ! -d /etc/haproxy ] ; then mkdir -p /etc/haproxy ; fi
# cp examples/haproxy.cfg /etc/haproxy/haproxy.cfg
# haproxy -v
exit 1
"""


version_cmd=''
download_url='http://www.rabbitmq.com/rabbitmq-signing-key-public.asc'
install_script="""
apt-key add rabbitmq-signing-key-public.asc
echo "deb http://www.rabbitmq.com/debian/ testing main" > /etc/apt/sources.list.d/rabbitmq.list
apt-get update
DEBIAN_FRONTEND=noninteractive apt-get install -y rabbitmq-server
"""


import glob
import imp
import importlib
import os
import re
import subprocess
import sys
import time
import urllib


import click


# We want this to persist across reboots so do not use /tmp
last_updated_file = '/var/lib/ifs/last_update'


def get_download_filename(url, target):
    """
    Given a URL and a path on the filesystem that may be a directory or file,
    return the filename that we will save our download into.

    E.g: http://blah/file.tar, /tmp -------------> /tmp/file.tar
    And: http://blah/file.tar, /tmp/archive.tar -> /tmp/archive.tar
    """
    if url and os.path.isdir(target):
        target += '/' + url.split('/')[-1]
        return target
    return None

def download(url, target):
    """
    Download URL to TARGET on your local filesystem. TARGET should be an
    absolute path to the location you want to download the file. Useful when
    curl / wget are not installed, but python is.
    """
    # Add progress bar via:
    # http://stackoverflow.com/a/22776/317916
    if not url:
        return None
    urllib.urlretrieve(url, target)
    return target

def list_app_names():
    path = os.path.dirname(os.path.realpath(__file__))
    files = glob.glob(path + '/source/*.py')
    results = []
    for f in files:
        i = os.path.basename(f).split('.')[0]
        if i not in ['__init__']:
            results.append(i)
    results.sort()
    return results

def list_apps():
    apps = []
    for app_name in list_app_names():
        apps.append(App.load(app_name))
    return apps

def search_app_strings(term):
    results = set()
    for app in list_apps():
        if app.name and term in app.name:
            results.add(app)
        if app.description and term in app.description:
            results.add(app)
    return results

def export_source(source):
    path = os.path.dirname(os.path.realpath(__file__))
    source_file = path + '/source/%s.py' % source
    if os.path.exists(source_file):
        return open(source_file).read()
    else:
        return None

def match_semver(string):
    return


def read_last_updated():
    if os.path.exists(last_updated_file):
        with open(last_updated_file) as f:
            try:
                return int(float(f.read()))
            except ValueError:
                return 0
    return 0


def write_last_updated():
    path = os.path.dirname(last_updated_file)
    if not os.path.isdir(path):
        os.makedirs(path, 0755)
    with open(last_updated_file, 'w') as f:
        f.write(str(time.time()))
        f.close()


def out_of_date(delay=86400):
    t = time.time()
    if t - read_last_updated() > delay:
        return True
    return False


class Cmd(object):
    returncode = None
    cmd = None
    output = None

    def __init__(self, cmd):
        cmd = "set -e\n%s" % cmd
        self.cmd = cmd

    def execute(self, autoprint=False):
        p = subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, bufsize=1, shell=True)
        output = ''
        for line in iter(p.stdout.readline, b''):
            output += line
            if autoprint:
                print line,
        p.communicate()
        self.output = output
        self.returncode = p.returncode
        return self.returncode == 0

    @classmethod
    def run(cls, cmd, autoprint=False):
        cmdo = cls(cmd)
        cmdo.execute(autoprint)
        return cmdo


class App(object):
    depends        = None
    description    = None
    download_url   = None
    install_script = None
    name           = None
    version        = None
    version_cmd    = None

    def __init__(self, name, module=None):
        self.name = name
        if module:
            for k, v in module.__dict__.iteritems():
                if k[0] != '_':
                    setattr(self, k, v)
        # self.__dict__.update(module.__dict__)

    @classmethod
    def load(cls, app_name):
        try:
            mod = importlib.import_module('ifs.source.%s' % app_name, '..source')
            app = cls(app_name, mod)
        except ImportError as e:
            app = None
        return app

    @classmethod
    def load_external(cls, path):
        d, f = os.path.split(path)
        name, ext = os.path.splitext(f)
        try:
            module = imp.load_source(name, path)
            app = cls(name, module)
            return app
        except Exception as e:
            return None

    def get_download_url(self, version=None):
        if not version:
            version = self.version
        if hasattr(self, 'download_url') and self.download_url:
            return self.download_url.replace('VERSION', version)
        else:
            return None

    def check_version(self):
        # Call self.version_cmd to check which version is currently installed
        cmd = Cmd.run(self.version_cmd)
        if hasattr(self, 'version_re'):
            r = re.compile(self.version_re)
        else:
            r = re.compile('(\d+\.\d+\.\d+)')
        matches = r.search(cmd.output)
        if matches is None:
            return None
        else:
            return matches.group(1)

    def info(self):
        info = {
            "description": self.description or self.name,
            "default_version": self.version or None,
            "current_version": self.check_version() or 'Not Installed',
            "dependencies": self.depends or None,
            "download_url": self.download_url or None,
        }
        return info

    def cmd_install_deps(self):
        if hasattr(self, 'depends') and self.depends:
            return 'apt-get install -y %s' % ' '.join(self.depends)
        else:
            return None

    def cmd_install_app(self, version=None):
        cmd = self.install_script
        if not version:
            version = self.version
        if version:
            cmd = cmd.replace('VERSION', version)
        return cmd

    def install(self, version=None, force=False):
        if not version:
            version = self.version

        # Create temp directory
        target='/tmp/ifs-%s-%s' % (self.name, self.version)
        if not os.path.exists(target):
            os.mkdir(target)
        os.chdir(target)

        # Download source
        dl_url = self.get_download_url(version)
        dl_file = get_download_filename(dl_url, target)
        if dl_file:
            if os.path.exists(dl_file):
                click.echo('Already downloaded %s' % dl_file)
            else:
                click.echo('Downloading %s to %s' % (dl_url, dl_file))
                if download(dl_url, dl_file):
                    click.echo('Download complete')

        # Install dependencies
        depc = self.cmd_install_deps()
        if depc:
            click.echo('Installing dependencies: %s' % depc)
            deps = Cmd.run(depc)
            if deps.returncode > 0:
                return deps

        click.echo('Installing from source')
        install = Cmd.run(self.cmd_install_app(version), True)
        return install



version_cmd='java -jar /usr/share/jenkins/jenkins.war --version'
version_re='(\d\.\d{3})'
depends=['wget']
install_script="""
wget -q -O - http://pkg.jenkins-ci.org/debian/jenkins-ci.org.key | apt-key add -
echo "deb http://pkg.jenkins-ci.org/debian binary/" > /etc/apt/sources.list.d/jenkins.list
apt-get update
apt-get install -y jenkins
"""


version='2.0.0'
version_cmd='riak version'
download_url='http://s3.amazonaws.com/downloads.basho.com/riak/2.0/2.0.0/ubuntu/trusty/riak_2.0.0-1_amd64.deb'
install_script="""
dpkg -i riak_2.0.0-1_amd64.deb
"""